@@ -38,6 +38,7 @@
 #include <cplusplus/Control.h>
 
 #include <utils/algorithm.h>
+#include <utils/porting.h>
 
 #include <QStack>
 #include <QHash>
@@ -60,15 +61,22 @@ static void addNames(const Name *name, QList<const Name *> *names, bool addAllNa
     }
 }
 
-static void path_helper(Symbol *symbol, QList<const Name *> *names)
+static void path_helper(Symbol *symbol,
+                        QList<const Name *> *names,
+                        LookupContext::InlineNamespacePolicy policy)
 {
     if (! symbol)
         return;
 
-    path_helper(symbol->enclosingScope(), names);
+    path_helper(symbol->enclosingScope(), names, policy);
 
     if (symbol->name()) {
         if (symbol->isClass() || symbol->isNamespace()) {
+            if (policy == LookupContext::HideInlineNamespaces) {
+                auto ns = symbol->asNamespace();
+                if (ns && ns->isInline())
+                    return;
+            }
             addNames(symbol->name(), names);
 
         } else if (symbol->isObjCClass() || symbol->isObjCBaseClass() || symbol->isObjCProtocol()
@@ -143,9 +151,9 @@ bool operator==(const FullyQualifiedName &left, const FullyQualifiedName &right)
     return compareFullyQualifiedName(left.fqn, right.fqn);
 }
 
-uint qHash(const FullyQualifiedName &fullyQualifiedName)
+Utils::QHashValueType qHash(const FullyQualifiedName &fullyQualifiedName)
 {
-    uint h = 0;
+    Utils::QHashValueType h = 0;
     for (int i = 0; i < fullyQualifiedName.fqn.size(); ++i) {
         if (const Name *n = fullyQualifiedName.fqn.at(i)) {
             if (const Identifier *id = n->identifier()) {
@@ -206,17 +214,18 @@ LookupContext &LookupContext::operator=(const LookupContext &other)
     return *this;
 }
 
-QList<const Name *> LookupContext::fullyQualifiedName(Symbol *symbol)
+QList<const Name *> LookupContext::fullyQualifiedName(Symbol *symbol, InlineNamespacePolicy policy)
 {
-    QList<const Name *> qualifiedName = path(symbol->enclosingScope());
+    QList<const Name *> qualifiedName = path(symbol->enclosingScope(), policy);
+    QList<const Name *> symbolNames;
     addNames(symbol->name(), &qualifiedName, /*add all names*/ true);
     return qualifiedName;
 }
 
-QList<const Name *> LookupContext::path(Symbol *symbol)
+QList<const Name *> LookupContext::path(Symbol *symbol, InlineNamespacePolicy policy)
 {
     QList<const Name *> names;
-    path_helper(symbol, &names);
+    path_helper(symbol, &names, policy);
     return names;
 }
 
@@ -232,19 +241,6 @@ static bool symbolIdentical(Symbol *s1, Symbol *s2)
     return QByteArray(s1->fileName()) == QByteArray(s2->fileName());
 }
 
-static const Name *toName(const QList<const Name *> &names, Control *control)
-{
-    const Name *n = nullptr;
-    for (int i = names.size() - 1; i >= 0; --i) {
-        if (! n)
-            n = names.at(i);
-        else
-            n = control->qualifiedNameId(names.at(i), n);
-    }
-
-    return n;
-}
-
 static bool isInlineNamespace(ClassOrNamespace *con, const Name *name)
 {
     const QList<LookupItem> items = con->find(name);
@@ -262,6 +258,27 @@ const Name *LookupContext::minimalName(Symbol *symbol, ClassOrNamespace *target,
 {
     const Name *n = nullptr;
     QList<const Name *> names = LookupContext::fullyQualifiedName(symbol);
+    ClassOrNamespace *current = target;
+
+    const auto getNameFromItems = [symbol, target, control](const QList<LookupItem> &items,
+            const QList<const Name *> &names) -> const Name * {
+        for (const LookupItem &item : items) {
+            if (!symbol->asUsingDeclaration() && !symbolIdentical(item.declaration(), symbol))
+                continue;
+
+            // eliminate inline namespaces
+            QList<const Name *> minimal = names;
+            for (int i = minimal.size() - 2; i >= 0; --i) {
+                const Name *candidate = control->toName(minimal.mid(0, i + 1));
+                if (isInlineNamespace(target, candidate))
+                    minimal.removeAt(i);
+            }
+
+            return control->toName(minimal);
+        }
+
+        return nullptr;
+    };
 
     for (int i = names.size() - 1; i >= 0; --i) {
         if (! n)
@@ -271,20 +288,24 @@ const Name *LookupContext::minimalName(Symbol *symbol, ClassOrNamespace *target,
 
         // once we're qualified enough to get the same symbol, break
         if (target) {
-            const QList<LookupItem> tresults = target->lookup(n);
-            foreach (const LookupItem &tr, tresults) {
-                if (symbolIdentical(tr.declaration(), symbol)) {
-                    // eliminate inline namespaces
-                    QList<const Name *> minimal = names.mid(i);
-                    for (int i = minimal.size() - 2; i >= 0; --i) {
-                        const Name *candidate = toName(minimal.mid(0, i + 1), control);
-                        if (isInlineNamespace(target, candidate))
-                            minimal.removeAt(i);
-                    }
-
-                    return toName(minimal, control);
+            const Name * const minimal = getNameFromItems(target->lookup(n), names.mid(i));
+            if (minimal)
+                return minimal;
+        }
+        if (current) {
+            const ClassOrNamespace * const nested = current->getNested(names.last());
+            if (nested) {
+                const QList<const Name *> nameList
+                        = names.mid(0, names.size() - i - 1) << names.last();
+                const QList<ClassOrNamespace *> usings = nested->usings();
+                for (ClassOrNamespace * const u : usings) {
+                    const Name * const minimal = getNameFromItems(u->lookup(symbol->name()),
+                                                                  nameList);
+                    if (minimal)
+                        return minimal;
                 }
             }
+            current = current->getNested(names.at(names.size() - i - 1));
         }
     }
 
@@ -758,10 +779,10 @@ void CreateBindings::lookupInScope(const Name *name, Scope *scope,
                                    const TemplateNameId *templateId,
                                    ClassOrNamespace *binding)
 {
-    if (! name) {
+    if (!name)
         return;
 
-    } else if (const OperatorNameId *op = name->asOperatorNameId()) {
+    if (const OperatorNameId *op = name->asOperatorNameId()) {
         for (Symbol *s = scope->find(op->kind()); s; s = s->next()) {
             if (! s->name())
                 continue;
@@ -779,8 +800,24 @@ void CreateBindings::lookupInScope(const Name *name, Scope *scope,
 
             result->append(item);
         }
+        return;
+    }
 
-    } else if (const Identifier *id = name->identifier()) {
+    if (const ConversionNameId * const conv = name->asConversionNameId()) {
+        if (Symbol * const s = scope->find(conv)) {
+            LookupItem item;
+            item.setDeclaration(s);
+            item.setBinding(binding);
+
+            if (Symbol *inst = instantiateTemplateFunction(name, s->asTemplate()))
+                item.setType(inst->type());
+
+            result->append(item);
+        }
+        return;
+    }
+
+    if (const Identifier *id = name->identifier()) {
         for (Symbol *s = scope->find(id); s; s = s->next()) {
             if (s->isFriend())
                 continue; // skip friends
@@ -899,6 +936,15 @@ ClassOrNamespace *ClassOrNamespace::findBlock(Block *block)
     return findBlock_helper(block, &processed, true);
 }
 
+ClassOrNamespace *ClassOrNamespace::getNested(const Name *name)
+{
+    flush();
+    const auto it = _classOrNamespaces.find(name);
+    if (it != _classOrNamespaces.cend())
+        return it->second;
+    return nullptr;
+}
+
 Symbol *ClassOrNamespace::lookupInScope(const QList<const Name *> &fullName)
 {
     if (!_scopeLookupCache) {
