@@ -29,7 +29,6 @@
 #include "modelnodepositionstorage.h"
 #include "abstractproperty.h"
 #include "bindingproperty.h"
-#include "enumeration.h"
 #include "filemanager/firstdefinitionfinder.h"
 #include "filemanager/objectlengthcalculator.h"
 #include "filemanager/qmlrefactoring.h"
@@ -42,22 +41,36 @@
 #include "rewriterview.h"
 #include "variantproperty.h"
 
+#include <enumeration.h>
+
+#ifndef QMLDESIGNER_TEST
+#include <projectexplorer/kit.h>
+#include <projectexplorer/session.h>
+#include <projectexplorer/target.h>
+
+#include <qtsupport/baseqtversion.h>
+#include <qtsupport/qtkitinformation.h>
+
+#endif
+
 #include <qmljs/qmljsevaluate.h>
 #include <qmljs/qmljslink.h>
 #include <qmljs/parser/qmljsast_p.h>
 #include <qmljs/qmljscheck.h>
 #include <qmljs/qmljsutils.h>
 #include <qmljs/qmljsmodelmanagerinterface.h>
-#include <qmljs/qmljsqrcparser.h>
 #include <qmljs/qmljsinterpreter.h>
 #include <qmljs/qmljsvalueowner.h>
 
+#include <utils/algorithm.h>
+#include <utils/qrcparser.h>
 #include <utils/qtcassert.h>
 
 #include <QSet>
 #include <QDir>
 #include <QLoggingCategory>
 #include <QRegularExpression>
+#include <QElapsedTimer>
 
 #include <memory>
 
@@ -75,9 +88,25 @@ bool isSupportedAttachedProperties(const QString &propertyName)
 
 QStringList supportedVersionsList()
 {
-    static const QStringList list = {
-        "2.0", "2.1", "2.2", "2.3", "2.4", "2.5", "2.6", "2.7", "2.8", "2.9", "2.10", "2.11", "2.12"
-    };
+    static const QStringList list = {"2.0",
+                                     "2.1",
+                                     "2.2",
+                                     "2.3",
+                                     "2.4",
+                                     "2.5",
+                                     "2.6",
+                                     "2.7",
+                                     "2.8",
+                                     "2.9",
+                                     "2.10",
+                                     "2.11",
+                                     "2.12",
+                                     "2.13",
+                                     "2.14",
+                                     "2.15",
+                                     "6.0",
+                                     "6.1",
+                                     "6.2"};
     return list;
 }
 
@@ -87,7 +116,12 @@ QStringList globalQtEnums()
         "Horizontal", "Vertical", "AlignVCenter", "AlignLeft", "LeftToRight", "RightToLeft",
         "AlignHCenter", "AlignRight", "AlignBottom", "AlignBaseline", "AlignTop", "BottomLeft",
         "LeftEdge", "RightEdge", "BottomEdge", "TopEdge", "TabFocus", "ClickFocus", "StrongFocus",
-        "WheelFocus", "NoFocus"
+        "WheelFocus", "NoFocus", "ArrowCursor", "UpArrowCursor", "CrossCursor", "WaitCursor",
+        "IBeamCursor", "SizeVerCursor", "SizeHorCursor", "SizeBDiagCursor", "SizeFDiagCursor",
+        "SizeAllCursor", "BlankCursor", "SplitVCursor", "SplitHCursor", "PointingHandCursor",
+        "ForbiddenCursor", "WhatsThisCursor", "BusyCursor", "OpenHandCursor", "ClosedHandCursor",
+        "DragCopyCursor", "DragMoveCursor", "DragLinkCursor", "TopToBottom",
+        "LeftButton", "RightButton", "MiddleButton", "BackButton", "ForwardButton", "AllButtons"
     };
 
     return list;
@@ -95,15 +129,26 @@ QStringList globalQtEnums()
 
 QStringList knownEnumScopes()
 {
-    static const QStringList list = {
-        "TextInput", "TextEdit", "Material", "Universal", "Font", "Shape", "ShapePath", "AbstractButton"
-    };
+    static const QStringList list = {"TextInput",
+                                     "TextEdit",
+                                     "Material",
+                                     "Universal",
+                                     "Font",
+                                     "Shape",
+                                     "ShapePath",
+                                     "AbstractButton",
+                                     "Text",
+                                     "ShaderEffectSource",
+                                     "Grid",
+                                     "ItemLayer",
+                                     "ImageLayer",
+                                     "SpriteLayer"};
     return list;
 }
 
 bool supportedQtQuickVersion(const QString &version)
 {
-    return supportedVersionsList().contains(version);
+    return version.isEmpty() || supportedVersionsList().contains(version);
 }
 
 QString stripQuotes(const QString &str)
@@ -278,19 +323,17 @@ bool isListElementType(const QmlDesigner::TypeName &type)
 
 bool isComponentType(const QmlDesigner::TypeName &type)
 {
-    return type == "Component"
-            || type == "Qt.Component"
-            || type == "QtQuick.Component"
-            || type == "<cpp>.QQmlComponent"
-            || type == "QQmlComponent";
+    return type == "Component" || type == "Qt.Component" || type == "QtQuick.Component"
+           || type == "QtQml.Component" || type == "<cpp>.QQmlComponent" || type == "QQmlComponent";
 }
 
 bool isCustomParserType(const QmlDesigner::TypeName &type)
 {
-    return type == "QtQuick.VisualItemModel" || type == "Qt.VisualItemModel" ||
-           type == "QtQuick.VisualDataModel" || type == "Qt.VisualDataModel" ||
-           type == "QtQuick.ListModel" || type == "Qt.ListModel" ||
-           type == "QtQuick.XmlListModel" || type == "Qt.XmlListModel";
+    return type == "QtQuick.VisualItemModel" || type == "Qt.VisualItemModel"
+           || type == "QtQuick.VisualDataModel" || type == "Qt.VisualDataModel"
+           || type == "QtQuick.ListModel" || type == "Qt.ListModel"
+           || type == "QtQml.Models.ListModel" || type == "QtQuick.XmlListModel"
+           || type == "Qt.XmlListModel";
 }
 
 
@@ -301,7 +344,7 @@ bool isPropertyChangesType(const QmlDesigner::TypeName &type)
 
 bool isConnectionsType(const QmlDesigner::TypeName &type)
 {
-    return type == "Connections" || type == "QtQuick.Connections" || type == "Qt.Connections";
+    return type == "Connections" || type == "QtQuick.Connections" || type == "Qt.Connections" || type == "QtQml.Connections";
 }
 
 bool propertyIsComponentType(const QmlDesigner::NodeAbstractProperty &property, const QmlDesigner::TypeName &type, QmlDesigner::Model *model)
@@ -350,8 +393,14 @@ bool compareJavaScriptExpression(const QString &expression1, const QString &expr
 }
 
 bool smartVeryFuzzyCompare(const QVariant &value1, const QVariant &value2)
-{ //we ignore slight changes on doubles and only check three digits
-    if ((value1.type() == QVariant::Double) || (value2.type() == QVariant::Double)) {
+{
+    //we ignore slight changes on doubles and only check three digits
+    const auto type1 = static_cast<QMetaType::Type>(value1.type());
+    const auto type2 = static_cast<QMetaType::Type>(value2.type());
+    if (type1 == QMetaType::Double
+            || type2 == QMetaType::Double
+            || type1 == QMetaType::Float
+            || type2 == QMetaType::Float) {
         bool ok1, ok2;
         qreal a = value1.toDouble(&ok1);
         qreal b = value2.toDouble(&ok2);
@@ -371,6 +420,13 @@ bool smartVeryFuzzyCompare(const QVariant &value1, const QVariant &value2)
     return false;
 }
 
+bool smartColorCompare(const QVariant &value1, const QVariant &value2)
+{
+    if ((value1.type() == QVariant::Color) || (value2.type() == QVariant::Color))
+        return value1.value<QColor>().rgba() == value2.value<QColor>().rgba();
+    return false;
+}
+
 bool equals(const QVariant &a, const QVariant &b)
 {
     if (a.canConvert<QmlDesigner::Enumeration>() && b.canConvert<QmlDesigner::Enumeration>())
@@ -379,6 +435,8 @@ bool equals(const QVariant &a, const QVariant &b)
         return true;
     if (smartVeryFuzzyCompare(a, b))
         return true;
+    if (smartColorCompare(a, b))
+        return true;
     return false;
 }
 
@@ -456,7 +514,7 @@ public:
                 if (!name.isEmpty() && name != QLatin1String("."))
                     typeName.prepend(name + QLatin1Char('.'));
             } else if (importInfo.isValid() && importInfo.type() == ImportType::QrcDirectory) {
-                QString path = QrcParser::normalizedQrcDirectoryPath(importInfo.path());
+                QString path = Utils::QrcParser::normalizedQrcDirectoryPath(importInfo.path());
                 path = path.mid(1, path.size() - ((path.size() > 1) ? 2 : 1));
                 const QString name = path.replace(QLatin1Char('/'), QLatin1Char('.'));
                 if (!name.isEmpty())
@@ -746,14 +804,16 @@ void TextToModelMerger::setupImports(const Document::Ptr &doc,
 {
     QList<Import> existingImports = m_rewriterView->model()->imports();
 
+    m_hasVersionlessImport = false;
+
     for (AST::UiHeaderItemList *iter = doc->qmlProgram()->headers; iter; iter = iter->next) {
         auto import = AST::cast<AST::UiImport *>(iter->headerItem);
         if (!import)
             continue;
 
         QString version;
-        if (import->versionToken.isValid())
-            version = textAt(doc, import->versionToken);
+        if (import->version != nullptr)
+            version = QString("%1.%2").arg(import->version->majorVersion).arg(import->version->minorVersion);
         const QString &as = import->importId.toString();
 
         if (!import->fileName.isEmpty()) {
@@ -765,13 +825,13 @@ void TextToModelMerger::setupImports(const Document::Ptr &doc,
                 differenceHandler.modelMissesImport(newImport);
         } else {
             QString importUri = toString(import->importUri);
-            if (importUri == QStringLiteral("Qt") && version == QStringLiteral("4.7")) {
-                importUri = QStringLiteral("QtQuick");
-                version = QStringLiteral("1.0");
-            }
+            if (version.isEmpty())
+                m_hasVersionlessImport = true;
 
-            const Import newImport =
-                    Import::createLibraryImport(importUri, version, as, m_rewriterView->importDirectories());
+            const Import newImport = Import::createLibraryImport(importUri,
+                                                                 version,
+                                                                 as,
+                                                                 m_rewriterView->importDirectories());
 
             if (!existingImports.removeOne(newImport))
                 differenceHandler.modelMissesImport(newImport);
@@ -817,7 +877,6 @@ static bool isBlacklistImport(const ImportKey &importKey, Model *model)
             || importKey.libraryQualifiedPath() == QStringLiteral("QtQuick.Dialogs")   //Unsupported
             || importKey.libraryQualifiedPath() == QStringLiteral("QtQuick.Controls.Styles")   //Unsupported
             || importKey.libraryQualifiedPath() == QStringLiteral("QtNfc") //Unsupported
-            || importKey.libraryQualifiedPath() == QStringLiteral("QtMultimedia")
             || importKey.libraryQualifiedPath() == QStringLiteral("Qt.WebSockets")
             || importKey.libraryQualifiedPath() == QStringLiteral("QtWebkit")
             || importKey.libraryQualifiedPath() == QStringLiteral("QtLocation")
@@ -827,8 +886,6 @@ static bool isBlacklistImport(const ImportKey &importKey, Model *model)
             || importKey.libraryQualifiedPath() == QStringLiteral("QtBluetooth")
             || importKey.libraryQualifiedPath() ==  QStringLiteral("Enginio")
 
-            // Don't show Quick X.X imports
-            || (importKey.splitPath.count() == 1 && importPathFirst == QStringLiteral("QtQuick"))
             || filterByMetaInfo(importKey, model);
 }
 
@@ -849,18 +906,38 @@ static void removeUsedImports(QHash<QString, ImportKey> &filteredPossibleImportK
         filteredPossibleImportKeys.remove(import.info.path());
 }
 
-static QList<QmlDesigner::Import> generatePossibleFileImports(const QString &path)
+static QList<QmlDesigner::Import> generatePossibleFileImports(const QString &path,
+                                                              const QList<QmlJS::Import> &usedImports)
 {
-    QList<QmlDesigner::Import> possibleImports;
+    QSet<QString> usedImportsSet;
+    for (const QmlJS::Import &i : usedImports)
+        usedImportsSet.insert(i.info.path());
 
-    foreach (const QString &subDir, QDir(path).entryList(QDir::Dirs | QDir::NoDot | QDir::NoDotDot)) {
-        QDir dir(path + "/" + subDir);
-        if (!dir.entryInfoList(QStringList("*.qml"), QDir::Files).isEmpty()
-                && dir.entryInfoList(QStringList("qmldir"), QDir::Files).isEmpty()) {
-            QmlDesigner::Import import = QmlDesigner::Import::createFileImport(subDir);
-            possibleImports.append(import);
+    QList<QmlDesigner::Import> possibleImports;
+    const QStringList qmlList("*.qml");
+    const QStringList qmldirList("qmldir");
+
+    QStringList fileImportPaths;
+    const QChar delimeter('/');
+
+    std::function<void(const QString &)> checkDir;
+    checkDir = [&](const QString &checkPath) {
+        const QStringList entries = QDir(checkPath).entryList(QDir::Dirs | QDir::NoDot | QDir::NoDotDot);
+        const QString checkPathDelim = checkPath + delimeter;
+        for (const QString &entry : entries) {
+            QDir dir(checkPathDelim + entry);
+            const QString dirPath = dir.path();
+            if (!dir.entryInfoList(qmlList, QDir::Files).isEmpty()
+                    && dir.entryInfoList(qmldirList, QDir::Files).isEmpty()
+                    && !usedImportsSet.contains(dirPath)) {
+                const QString importName = dir.path().mid(path.size() + 1);
+                QmlDesigner::Import import = QmlDesigner::Import::createFileImport(importName);
+                possibleImports.append(import);
+            }
+            checkDir(dirPath);
         }
-    }
+    };
+    checkDir(path);
 
     return possibleImports;
 }
@@ -868,24 +945,53 @@ static QList<QmlDesigner::Import> generatePossibleFileImports(const QString &pat
 static QList<QmlDesigner::Import> generatePossibleLibraryImports(const QHash<QString, ImportKey> &filteredPossibleImportKeys)
 {
     QList<QmlDesigner::Import> possibleImports;
+    QSet<QString> controlsImplVersions;
+    bool hasVersionedControls = false;
+    bool hasVersionlessControls = false;
+    const QString controlsName = "QtQuick.Controls";
+    const QString controlsImplName = "QtQuick.Controls.impl";
 
-    foreach (const ImportKey &importKey, filteredPossibleImportKeys) {
+    for (const ImportKey &importKey : filteredPossibleImportKeys) {
         QString libraryName = importKey.splitPath.join(QLatin1Char('.'));
         int majorVersion = importKey.majorVersion;
         if (majorVersion >= 0) {
             int minorVersion = (importKey.minorVersion == LanguageUtils::ComponentVersion::NoVersion) ? 0 : importKey.minorVersion;
             QString version = QStringLiteral("%1.%2").arg(majorVersion).arg(minorVersion);
-            possibleImports.append(QmlDesigner::Import::createLibraryImport(libraryName, version));
+            if (!libraryName.endsWith(".impl"))
+                possibleImports.append(QmlDesigner::Import::createLibraryImport(libraryName, version));
+
+            // In Qt6, QtQuick.Controls itself doesn't have any version as it has no types,
+            // so it never gets added normally to possible imports.
+            // We work around this by injecting corresponding QtQuick.Controls version for each
+            // found impl version, if no valid QtQuick.Controls versions are found.
+            if (!hasVersionedControls) {
+                if (libraryName == controlsImplName)
+                    controlsImplVersions.insert(version);
+                else if (libraryName == controlsName)
+                    hasVersionedControls = true;
+            }
+        } else if (!hasVersionlessControls && libraryName == controlsName) {
+            // If QtQuick.Controls module is not included even in non-versioned, it means
+            // QtQuick.Controls is either in use or not available at all,
+            // so we shouldn't inject it.
+            hasVersionlessControls = true;
         }
     }
 
+    if (hasVersionlessControls && !hasVersionedControls && !controlsImplVersions.isEmpty()) {
+        for (const auto &version : std::as_const(controlsImplVersions))
+            possibleImports.append(QmlDesigner::Import::createLibraryImport(controlsName, version));
+    }
+
+
     return possibleImports;
 }
 
 void TextToModelMerger::setupPossibleImports(const QmlJS::Snapshot &snapshot, const QmlJS::ViewerContext &viewContext)
 {
-    QHash<QString, ImportKey> filteredPossibleImportKeys =
-            filterPossibleImportKeys(snapshot.importDependencies()->libraryImports(viewContext), m_rewriterView->model());
+    m_possibleImportKeys = snapshot.importDependencies()->libraryImports(viewContext);
+    QHash<QString, ImportKey> filteredPossibleImportKeys
+        = filterPossibleImportKeys(m_possibleImportKeys, m_rewriterView->model());
 
     const QmlJS::Imports *imports = m_scopeChain->context()->imports(m_document.data());
     if (imports)
@@ -893,7 +999,7 @@ void TextToModelMerger::setupPossibleImports(const QmlJS::Snapshot &snapshot, co
 
     QList<QmlDesigner::Import> possibleImports = generatePossibleLibraryImports(filteredPossibleImportKeys);
 
-    possibleImports.append(generatePossibleFileImports(document()->path()));
+    possibleImports.append(generatePossibleFileImports(document()->path(), imports->all()));
 
     if (m_rewriterView->isAttached())
         m_rewriterView->model()->setPossibleImports(possibleImports);
@@ -907,15 +1013,29 @@ void TextToModelMerger::setupUsedImports()
 
      const QList<QmlJS::Import> allImports = imports->all();
 
+     QSet<QString> usedImportsSet;
      QList<Import> usedImports;
 
-     foreach (const QmlJS::Import &import, allImports) {
-         if (import.used && !import.info.name().isEmpty()) {
-            if (import.info.type() == ImportType::Library) {
-                usedImports.append(Import::createLibraryImport(import.info.name(), import.info.version().toString(), import.info.as()));
-            } else if (import.info.type() == ImportType::Directory || import.info.type() == ImportType::File) {
+     // populate usedImportsSet from current model nodes
+     const QList<ModelNode> allModelNodes = m_rewriterView->allModelNodes();
+     for (const ModelNode &modelNode : allModelNodes) {
+         QString type = QString::fromUtf8(modelNode.type());
+         if (type.contains('.'))
+             usedImportsSet.insert(type.left(type.lastIndexOf('.')));
+     }
+
+     for (const QmlJS::Import &import : allImports) {
+         QString version = import.info.version().toString();
+
+         if (!import.info.version().isValid())
+             version = getHighestPossibleImport(import.info.name());
+
+         if (!import.info.name().isEmpty() && usedImportsSet.contains(import.info.name())) {
+            if (import.info.type() == ImportType::Library)
+                usedImports.append(
+                    Import::createLibraryImport(import.info.name(), version, import.info.as()));
+            else if (import.info.type() == ImportType::Directory || import.info.type() == ImportType::File)
                 usedImports.append(Import::createFileImport(import.info.name(), import.info.version().toString(), import.info.as()));
-            }
          }
      }
 
@@ -925,6 +1045,15 @@ void TextToModelMerger::setupUsedImports()
 
 Document::MutablePtr TextToModelMerger::createParsedDocument(const QUrl &url, const QString &data, QList<DocumentMessage> *errors)
 {
+    if (data.isEmpty()) {
+        if (errors) {
+            QmlJS::DiagnosticMessage msg;
+            msg.message = QObject::tr("Empty document");
+            errors->append(DocumentMessage(msg, url));
+        }
+        return {};
+    }
+
     const QString fileName = url.toLocalFile();
 
     Dialect dialect = ModelManagerInterface::guessLanguageOfFile(fileName);
@@ -951,15 +1080,12 @@ bool TextToModelMerger::load(const QString &data, DifferenceHandler &differenceH
 {
     qCInfo(rewriterBenchmark) << Q_FUNC_INFO;
 
-    const bool justSanityCheck = !differenceHandler.isValidator();
+    const bool justSanityCheck = !differenceHandler.isAmender();
 
-    QTime time;
+    QElapsedTimer time;
     if (rewriterBenchmark().isInfoEnabled())
         time.start();
 
-    // maybe the project environment (kit, ...) changed, so we need to clean old caches
-    NodeMetaInfo::clearCache();
-
     const QUrl url = m_rewriterView->model()->fileUrl();
 
     m_qrcMapping.clear();
@@ -969,8 +1095,11 @@ bool TextToModelMerger::load(const QString &data, DifferenceHandler &differenceH
     setActive(true);
     m_rewriterView->setIncompleteTypeInformation(false);
 
+    // maybe the project environment (kit, ...) changed, so we need to clean old caches
+    m_rewriterView->model()->clearMetaInfoCache();
+
     try {
-        Snapshot snapshot = m_rewriterView->textModifier()->qmljsSnapshot();
+        Snapshot snapshot = TextModifier::qmljsSnapshot();
 
         QList<DocumentMessage> errors;
         QList<DocumentMessage> warnings;
@@ -994,7 +1123,7 @@ bool TextToModelMerger::load(const QString &data, DifferenceHandler &differenceH
         }
 
 
-        m_vContext = ModelManagerInterface::instance()->defaultVContext(Dialect::Qml, m_document);
+        m_vContext = ModelManagerInterface::instance()->projectVContext(Dialect::Qml, m_document);
         ReadingContext ctxt(snapshot, m_document, m_vContext);
         m_scopeChain = QSharedPointer<const ScopeChain>(
                     new ScopeChain(ctxt.scopeChain()));
@@ -1002,8 +1131,10 @@ bool TextToModelMerger::load(const QString &data, DifferenceHandler &differenceH
         qCInfo(rewriterBenchmark) << "linked:" << time.elapsed();
         collectLinkErrors(&errors, ctxt);
 
-        setupImports(m_document, differenceHandler);
         setupPossibleImports(snapshot, m_vContext);
+        setupImports(m_document, differenceHandler);
+
+        qCInfo(rewriterBenchmark) << "imports setup:" << time.elapsed();
 
         collectImportErrors(&errors);
 
@@ -1019,7 +1150,6 @@ bool TextToModelMerger::load(const QString &data, DifferenceHandler &differenceH
                 m_rewriterView->setWarnings(warnings);
             qCInfo(rewriterBenchmark) << "checked semantic errors:" << time.elapsed();
         }
-        setupUsedImports();
 
         AST::UiObjectMember *astRootNode = nullptr;
         if (AST::UiProgram *program = m_document->qmlProgram())
@@ -1031,6 +1161,8 @@ bool TextToModelMerger::load(const QString &data, DifferenceHandler &differenceH
 
         qCInfo(rewriterBenchmark) << "synced nodes:" << time.elapsed();
 
+        setupUsedImports();
+
         setActive(false);
 
         return true;
@@ -1078,7 +1210,7 @@ void TextToModelMerger::syncNode(ModelNode &modelNode,
         return;
     }
 
-    if (modelNode.isRootNode() && isComponentType(typeName)) {
+    if (modelNode.isRootNode() && !m_rewriterView->allowComponentRoot() && isComponentType(typeName)) {
         for (AST::UiObjectMemberList *iter = astInitializer->members; iter; iter = iter->next) {
             if (auto def = AST::cast<AST::UiObjectDefinition *>(iter->member)) {
                 syncNode(modelNode, def, context, differenceHandler);
@@ -1097,19 +1229,28 @@ void TextToModelMerger::syncNode(ModelNode &modelNode,
         differenceHandler.typeDiffers(isRootNode, modelNode, typeName,
                                       majorVersion, minorVersion,
                                       astNode, context);
-        if (!isRootNode)
+
+        if (!modelNode.isValid())
+            return;
+
+        if (!isRootNode && modelNode.majorVersion() != -1 && modelNode.minorVersion() != -1) {
+            qWarning() << "Preempting Node sync. Type differs" << modelNode <<
+                          modelNode.majorVersion() << modelNode.minorVersion();
             return; // the difference handler will create a new node, so we're done.
+        }
     }
 
     if (isComponentType(typeName) || isImplicitComponent)
-        setupComponentDelayed(modelNode, differenceHandler.isValidator());
+        setupComponentDelayed(modelNode, differenceHandler.isAmender());
+    else if (isCustomParserType(typeName))
+        setupCustomParserNodeDelayed(modelNode, differenceHandler.isAmender());
+    else if (!modelNode.nodeSource().isEmpty() || modelNode.nodeSourceType() != ModelNode::NodeWithoutSource)
+        clearImplicitComponentDelayed(modelNode, differenceHandler.isAmender());
 
-    if (isCustomParserType(typeName))
-        setupCustomParserNodeDelayed(modelNode, differenceHandler.isValidator());
 
     context->enterScope(astNode);
 
-    QSet<PropertyName> modelPropertyNames = QSet<PropertyName>::fromList(modelNode.propertyNames());
+    QSet<PropertyName> modelPropertyNames = Utils::toSet(modelNode.propertyNames());
     if (!modelNode.id().isEmpty())
         modelPropertyNames.insert("id");
     QList<AST::UiObjectMember *> defaultPropertyItems;
@@ -1178,10 +1319,7 @@ void TextToModelMerger::syncNode(ModelNode &modelNode,
             if (property->type == AST::UiPublicMember::Signal)
                 continue; // QML designer doesn't support this yet.
 
-            if (property->name.isEmpty() || !property->isValid())
-                continue; // better safe than sorry.
-
-            const QStringRef astName = property->name;
+            const QStringView astName = property->name;
             QString astValue;
             if (property->statement)
                 astValue = textAt(context->doc(),
@@ -1193,7 +1331,7 @@ void TextToModelMerger::syncNode(ModelNode &modelNode,
                 astValue = astValue.left(astValue.length() - 1);
             astValue = astValue.trimmed();
 
-            const TypeName &astType = property->memberTypeName().toUtf8();
+            const TypeName &astType = property->memberType->name.toUtf8();
             AbstractProperty modelProperty = modelNode.property(astName.toUtf8());
 
             if (property->binding) {
@@ -1215,7 +1353,7 @@ void TextToModelMerger::syncNode(ModelNode &modelNode,
 
     if (!defaultPropertyItems.isEmpty()) {
         if (isComponentType(modelNode.type()))
-            setupComponentDelayed(modelNode, differenceHandler.isValidator());
+            setupComponentDelayed(modelNode, differenceHandler.isAmender());
         if (defaultPropertyName.isEmpty()) {
             qWarning() << "No default property for node type" << modelNode.type() << ", ignoring child items.";
         } else {
@@ -1250,12 +1388,12 @@ static QVariant parsePropertyExpression(AST::ExpressionNode *expressionNode)
 {
     Q_ASSERT(expressionNode);
 
-    auto arrayLiteral = AST::cast<AST::ArrayLiteral *>(expressionNode);
+    auto arrayLiteral = AST::cast<AST::ArrayPattern *>(expressionNode);
 
     if (arrayLiteral) {
         QList<QVariant> variantList;
-        for (AST::ElementList *it = arrayLiteral->elements; it; it = it->next)
-            variantList << parsePropertyExpression(it->expression);
+        for (AST::PatternElementList *it = arrayLiteral->elements; it; it = it->next)
+            variantList << parsePropertyExpression(it->element->initializer);
         return variantList;
     }
 
@@ -1642,7 +1780,7 @@ void ModelValidator::signalHandlerSourceDiffer(SignalHandlerProperty &modelPrope
 {
     Q_UNUSED(modelProperty)
     Q_UNUSED(javascript)
-    QTC_ASSERT(modelProperty.source() == javascript, return);
+    QTC_ASSERT(compareJavaScriptExpression(modelProperty.source(), javascript), return);
 }
 
 void ModelValidator::shouldBeSignalHandlerProperty(AbstractProperty &modelProperty, const QString & /*javascript*/)
@@ -1736,6 +1874,11 @@ void ModelValidator::typeDiffers(bool /*isRootNode*/,
         qDebug() << typeName << modelNode.majorVersion() << majorVersion;
     }
 
+    if (modelNode.minorVersion() != minorVersion) {
+        qDebug() << Q_FUNC_INFO << modelNode;
+        qDebug() << typeName << modelNode.minorVersion() << minorVersion;
+    }
+
     QTC_ASSERT(modelNode.majorVersion() == majorVersion, return);
     QTC_ASSERT(modelNode.minorVersion() == minorVersion, return);
     QTC_ASSERT(0, return);
@@ -1989,16 +2132,21 @@ void TextToModelMerger::setupComponent(const ModelNode &node)
 
     QString componentText = m_rewriterView->extractText({node}).value(node);
 
-    if (componentText.isEmpty())
+    if (componentText.isEmpty() && node.nodeSource().isEmpty())
         return;
 
     QString result = extractComponentFromQml(componentText);
 
-    if (result.isEmpty())
+    if (result.isEmpty() && node.nodeSource().isEmpty())
         return; //No object definition found
 
     if (node.nodeSource() != result)
-        ModelNode(node).setNodeSource(result);
+        ModelNode(node).setNodeSource(result, ModelNode::NodeWithComponentSource);
+}
+
+void TextToModelMerger::clearImplicitComponent(const ModelNode &node)
+{
+    ModelNode(node).setNodeSource({}, ModelNode::NodeWithoutSource);
 }
 
 void TextToModelMerger::collectLinkErrors(QList<DocumentMessage> *errors, const ReadingContext &ctxt)
@@ -2014,7 +2162,7 @@ void TextToModelMerger::collectLinkErrors(QList<DocumentMessage> *errors, const
 void TextToModelMerger::collectImportErrors(QList<DocumentMessage> *errors)
 {
     if (m_rewriterView->model()->imports().isEmpty()) {
-        const QmlJS::DiagnosticMessage diagnosticMessage(QmlJS::Severity::Error, AST::SourceLocation(0, 0, 0, 0), QCoreApplication::translate("QmlDesigner::TextToModelMerger", "No import statements found"));
+        const QmlJS::DiagnosticMessage diagnosticMessage(QmlJS::Severity::Error, SourceLocation(0, 0, 0, 0), QCoreApplication::translate("QmlDesigner::TextToModelMerger", "No import statements found."));
         errors->append(DocumentMessage(diagnosticMessage, QUrl::fromLocalFile(m_document->fileName())));
     }
 
@@ -2024,10 +2172,49 @@ void TextToModelMerger::collectImportErrors(QList<DocumentMessage> *errors)
 
             if (supportedQtQuickVersion(import.version())) {
                 hasQtQuick = true;
+
+#ifndef QMLDESIGNER_TEST
+                auto target = ProjectExplorer::SessionManager::startupTarget();
+                if (target) {
+                    QtSupport::BaseQtVersion *currentQtVersion = QtSupport::QtKitAspect::qtVersion(
+                        target->kit());
+                    if (currentQtVersion && currentQtVersion->isValid()) {
+                        const bool qt6import = import.version().startsWith("6");
+
+                        if (currentQtVersion->qtVersion().majorVersion == 5
+                            && (m_hasVersionlessImport || qt6import)) {
+                            const QmlJS::DiagnosticMessage diagnosticMessage(
+                                QmlJS::Severity::Error,
+                                SourceLocation(0, 0, 0, 0),
+                                QCoreApplication::translate(
+                                    "QmlDesigner::TextToModelMerger",
+                                    "Qt Quick 6 is not supported with a Qt 5 kit."));
+                            errors->prepend(
+                                DocumentMessage(diagnosticMessage,
+                                                QUrl::fromLocalFile(m_document->fileName())));
+                        }
+                    } else {
+                        const QmlJS::DiagnosticMessage
+                            diagnosticMessage(QmlJS::Severity::Error,
+                                              SourceLocation(0, 0, 0, 0),
+                                              QCoreApplication::translate(
+                                                  "QmlDesigner::TextToModelMerger",
+                                                  "The Design Mode requires a valid Qt kit."));
+                        errors->prepend(
+                            DocumentMessage(diagnosticMessage,
+                                            QUrl::fromLocalFile(m_document->fileName())));
+                    }
+                }
+#endif
+
             } else {
-                const QmlJS::DiagnosticMessage diagnosticMessage(QmlJS::Severity::Error, AST::SourceLocation(0, 0, 0, 0),
-                                                                 QCoreApplication::translate("QmlDesigner::TextToModelMerger", "Unsupported QtQuick version"));
-                errors->append(DocumentMessage(diagnosticMessage, QUrl::fromLocalFile(m_document->fileName())));
+                const QmlJS::DiagnosticMessage
+                    diagnosticMessage(QmlJS::Severity::Error,
+                                      SourceLocation(0, 0, 0, 0),
+                                      QCoreApplication::translate("QmlDesigner::TextToModelMerger",
+                                                                  "Unsupported Qt Quick version."));
+                errors->append(DocumentMessage(diagnosticMessage,
+                                               QUrl::fromLocalFile(m_document->fileName())));
             }
         }
     }
@@ -2064,6 +2251,18 @@ void TextToModelMerger::collectSemanticErrorsAndWarnings(QList<DocumentMessage>
         if (message.severity == Severity::Warning)
             warnings->append(DocumentMessage(message.toDiagnosticMessage(), fileNameUrl));
     }
+
+    for (const Import &import : m_rewriterView->model()->imports()) {
+        if (import.isLibraryImport() && import.url() == "QtQuick3D") {
+            const QString version = getHighestPossibleImport(import.url());
+            if (!import.version().isEmpty() && Import::majorFromVersion(version) > import.majorVersion()) {
+                errors->append(DocumentMessage(
+                    QObject::tr(
+                        "The selected version of the Qt Quick 3D module is not supported with the selected Qt version.")
+                        .arg(version)));
+            }
+        }
+    }
 }
 
 void TextToModelMerger::populateQrcMapping(const QString &filePath)
@@ -2096,9 +2295,9 @@ void TextToModelMerger::addIsoIconQrcMapping(const QUrl &fileUrl)
     } while (dir.cdUp());
 }
 
-void TextToModelMerger::setupComponentDelayed(const ModelNode &node, bool synchron)
+void TextToModelMerger::setupComponentDelayed(const ModelNode &node, bool synchronous)
 {
-    if (synchron) {
+    if (synchronous) {
         setupComponent(node);
     } else {
         m_setupComponentList.insert(node);
@@ -2113,19 +2312,19 @@ void TextToModelMerger::setupCustomParserNode(const ModelNode &node)
 
     QString modelText = m_rewriterView->extractText({node}).value(node);
 
-    if (modelText.isEmpty())
+    if (modelText.isEmpty() && node.nodeSource().isEmpty())
         return;
 
     if (node.nodeSource() != modelText)
-        ModelNode(node).setNodeSource(modelText);
+        ModelNode(node).setNodeSource(modelText, ModelNode::NodeWithCustomParserSource);
 
 }
 
-void TextToModelMerger::setupCustomParserNodeDelayed(const ModelNode &node, bool synchron)
+void TextToModelMerger::setupCustomParserNodeDelayed(const ModelNode &node, bool synchronous)
 {
     Q_ASSERT(isCustomParserType(node.type()));
 
-    if (synchron) {
+    if (synchronous) {
         setupCustomParserNode(node);
     } else {
         m_setupCustomParserList.insert(node);
@@ -2133,15 +2332,32 @@ void TextToModelMerger::setupCustomParserNodeDelayed(const ModelNode &node, bool
     }
 }
 
+void TextToModelMerger::clearImplicitComponentDelayed(const ModelNode &node, bool synchronous)
+{
+    Q_ASSERT(!isComponentType(node.type()));
+
+    if (synchronous) {
+        clearImplicitComponent(node);
+    } else {
+        m_clearImplicitComponentList.insert(node);
+        m_setupTimer.start();
+    }
+}
+
 void TextToModelMerger::delayedSetup()
 {
-    foreach (const ModelNode node, m_setupComponentList)
+    for (const ModelNode &node : std::as_const(m_setupComponentList))
         setupComponent(node);
 
-    foreach (const ModelNode node, m_setupCustomParserList)
+    for (const ModelNode &node : std::as_const(m_setupCustomParserList))
         setupCustomParserNode(node);
+
+    for (const ModelNode &node : std::as_const(m_clearImplicitComponentList))
+        clearImplicitComponent(node);
+
     m_setupCustomParserList.clear();
     m_setupComponentList.clear();
+    m_clearImplicitComponentList.clear();
 }
 
 QSet<QPair<QString, QString> > TextToModelMerger::qrcMapping() const
@@ -2149,15 +2365,65 @@ QSet<QPair<QString, QString> > TextToModelMerger::qrcMapping() const
     return m_qrcMapping;
 }
 
+QList<QmlTypeData> TextToModelMerger::getQMLSingletons() const
+{
+    QList<QmlTypeData> list;
+    if (!m_scopeChain || !m_scopeChain->document())
+        return list;
+
+    const QmlJS::Imports *imports = m_scopeChain->context()->imports(
+        m_scopeChain->document().data());
+
+    if (!imports)
+        return list;
+
+    for (const QmlJS::Import &import : imports->all()) {
+        if (import.info.type() == ImportType::Library && !import.libraryPath.isEmpty()) {
+            const LibraryInfo &libraryInfo = m_scopeChain->context()->snapshot().libraryInfo(
+                import.libraryPath);
+
+            for (const QmlDirParser::Component &component : libraryInfo.components()) {
+                if (component.singleton) {
+                    QmlTypeData qmlData;
+
+                    qmlData.typeName = component.typeName;
+                    qmlData.importUrl = import.info.name();
+                    qmlData.versionString = import.info.version().toString();
+                    qmlData.isSingleton = component.singleton;
+
+                    list.append(qmlData);
+                }
+            }
+        }
+    }
+    return list;
+}
+
 QString TextToModelMerger::textAt(const Document::Ptr &doc,
-                                  const AST::SourceLocation &location)
+                                  const SourceLocation &location)
 {
     return doc->source().mid(location.offset, location.length);
 }
 
 QString TextToModelMerger::textAt(const Document::Ptr &doc,
-                                  const AST::SourceLocation &from,
-                                  const AST::SourceLocation &to)
+                                  const SourceLocation &from,
+                                  const SourceLocation &to)
 {
     return doc->source().mid(from.offset, to.end() - from.begin());
 }
+
+QString TextToModelMerger::getHighestPossibleImport(const QString &importName) const
+{
+    QString version = "2.15";
+    int maj = -1;
+    const auto imports = m_possibleImportKeys.values();
+    for (const ImportKey &import : imports) {
+        if (importName == import.libraryQualifiedPath()) {
+            if (import.majorVersion > maj) {
+                version = QString("%1.%2").arg(import.majorVersion).arg(import.minorVersion);
+                maj = import.majorVersion;
+            }
+        }
+    }
+    return version;
+}
